/*
 * cp_pci.c
 *
 * KSI Labs, LLC CENTIPEDE driver module.
 *
 * Author: Sergey Kubushyn <ksi@koi8.net>
 *
 * Copyright (C) 2010 Sergey Kubushyn, KSI Labs, LLC
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/cdev.h>
#include <linux/pci.h>
#include <linux/wait.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/spinlock.h>
#include <linux/pid.h>

#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/siginfo.h>
#include <asm/signal.h>

#include "centipede.h"

#undef DEBUG

#define MAX_BOARDS	8		/* maximum of pc board possible */
#define IOMEM_LEN	4096		/* IOMEM is PAGE_SIZE long */

#ifndef PCI_VENDOR_ID_KSILABS
#define PCI_VENDOR_ID_KSILABS			0x4150	/* Change when got ID */
#define PCI_DEVICE_ID_KSILABS_CENTIPEDE		0x0007
#endif

#define CP_MOD		"centipede.ko: "

static struct pci_device_id cp_pci_tbl[] = {
	{ PCI_VDEVICE(KSILABS, PCI_DEVICE_ID_KSILABS_CENTIPEDE) },
	{ 0 }
};
MODULE_DEVICE_TABLE(pci, cp_pci_tbl);

MODULE_AUTHOR("Sergey Kubushyn <ksi@koi8.net>");
MODULE_DESCRIPTION("Driver for KSI Labs CENTIPEDE board");
MODULE_LICENSE("GPL");

MODULE_SUPPORTED_DEVICE("centipede");

static int cp_major;

struct cp_board {
	unsigned long		phys_io;
	void __iomem		*virt_io;
	unsigned int		irq;
	unsigned long		irq_st;
	unsigned long		irq_st2;
	char			cp_name[16];
	struct device		*cp_dev;
	struct fasync_struct	*async_queue;
	struct cdev		cdev;
	spinlock_t		mutex;
	char			reason;
};

static struct cp_board *cp_brds[MAX_BOARDS];

static struct class *cp_class;

static int cp_fasync(int fd, struct file *filp, int mode)
{
	struct cp_board	*dev = filp->private_data;
	
	return fasync_helper(fd, filp, mode, &dev->async_queue);
}

static ssize_t cp_read (struct file *filp, char __user *buf, size_t len, loff_t *offset)
{
	struct cp_board	*dev = filp->private_data;

	if (copy_to_user(buf, &dev->reason, 1))
		return -EFAULT;

	return 1;
}

static ssize_t cp_write (struct file *filp, const char __user *buf, size_t len, loff_t *offset)
{
	struct cp_board	*dev = filp->private_data;
	char		tmp;
	unsigned long	flags;
	
	if (copy_from_user(&tmp, buf, 1))
		return -EFAULT;

	spin_lock_irqsave(&dev->mutex, flags);
	dev->reason &= ~tmp;
	spin_unlock_irqrestore(&dev->mutex, flags);
	
	return len;
}

static int cp_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
	struct cp_board	*dev = filp->private_data;
    
	if (_IOC_TYPE(cmd) != CP_IOC_MAGIC) return -ENOTTY;
	if (_IOC_NR(cmd) > CP_IOC_MAXNR) return -ENOTTY;

#if 0	/* We are only passing values so far so leave this for later */
	if (_IOC_DIR(cmd) & _IOC_READ)
		err = !access_ok(VERIFY_WRITE, (void __user *)arg, _IOC_SIZE(cmd));
	else if (_IOC_DIR(cmd) & _IOC_WRITE)
		err =  !access_ok(VERIFY_READ, (void __user *)arg, _IOC_SIZE(cmd));
	if (err) return -EFAULT;
#endif

	switch(cmd) {
		case CP_IOCHOOKIRQ:
			break;
		case CP_IOCDROPIRQ:
			break;
		default:
			return -ENOTTY;
	}

	return 0;
}

#define IRQ_DAC_MASK	(CP_IRQ_DAC_A | CP_IRQ_DAC_B)
#define IRQ_ADC_MASK	(CP_IRQ_ADC_0 | CP_IRQ_ADC_1 | CP_IRQ_ADC_1 | CP_IRQ_ADC_3)

static irqreturn_t cp_interrupt(int irq, void *dev_id)
{
	struct cp_board	*dev = dev_id;
	
	dev->irq_st |= readl(dev->virt_io + CP_IRQSTAT_OFF);
	dev->irq_st2 |= readl(dev->virt_io + CP_IRQSTAT2_OFF) & ~CP_BRK_STATUS;

	writel(dev->irq_st, dev->virt_io + CP_IRQSTAT_RAW_OFF);
	writel(dev->irq_st2, dev->virt_io + CP_IRQSTAT2_RAW_OFF);

	spin_lock(&dev->mutex);

	if (!dev->irq_st)
		dev->reason |= IRQ_DIN;
		
	if (!(dev->irq_st2 & IRQ_DAC_MASK))
		dev->reason |= IRQ_DAC;

	if (!(dev->irq_st2 & IRQ_ADC_MASK))
		dev->reason |= IRQ_ADC;

	if (!(dev->irq_st2 & CP_IRQ_BRKLOST))
		dev->reason |= IRQ_BRKLOST;

	spin_unlock(&dev->mutex);

	if (dev->async_queue)
		kill_fasync(&dev->async_queue, SIGIO, POLL_MSG);

	return IRQ_HANDLED;
}

static int cp_open(struct inode *inode, struct file *filp)
{
	struct cp_board	*dev;

	dev = container_of(inode->i_cdev, struct cp_board, cdev);

	filp->private_data = dev;
	
	return nonseekable_open(inode, filp);
}

static int cp_release(struct inode *inode, struct file *filp)
{
	cp_fasync(-1, filp, 0);
	
	return 0;	
}

static int cp_mmap(struct file *filp, struct vm_area_struct *vma)
{
	struct cp_board	*brd = filp->private_data;
	unsigned long	off = vma->vm_pgoff << PAGE_SHIFT;
	unsigned long 	pfn = brd->phys_io >> PAGE_SHIFT;
	unsigned long	vsize = vma->vm_end - vma->vm_start;
	unsigned long	psize = IOMEM_LEN - off;

	if (vsize > psize)
		return -EINVAL;

	if (remap_pfn_range(vma, vma->vm_start, pfn, vsize, vma->vm_page_prot))
		return -EAGAIN;

	return 0;
}

#define cp_reg_show(reg,off)	\
static ssize_t			\
cp_##reg##_show(struct device *dev, struct device_attribute *attr, char *buf)	\
{				\
	struct cp_board *brd = dev_get_drvdata(dev);	\
	return sprintf(buf, "0x%x", readl(brd->virt_io + CP_##off##_OFF));	\
}

#define cp_reg_write(reg,off)	\
static ssize_t			\
cp_##reg##_write(struct device *dev, struct device_attribute *attr,	\
	const char *buf, size_t count)	\
{				\
	struct cp_board *brd = dev_get_drvdata(dev);	\
	unsigned long val = simple_strtoul(buf, NULL, 16);	\
	unsigned long flags;	\
	\
	spin_lock_irqsave(&brd->mutex, flags);	\
	writel(val, brd->virt_io + CP_##off##_OFF);	\
	spin_unlock_irqrestore(&brd->mutex, flags);	\
	\
	return count;	\
}

cp_reg_show(dout, DOUT);
cp_reg_write(dout, DOUT);
cp_reg_write(dout_set, DOUT_SET);
cp_reg_write(dout_clr, DOUT_CLR);
cp_reg_show(din, DIN);
cp_reg_show(debounce, DEBOUNCE);
cp_reg_write(debounce, DEBOUNCE);
cp_reg_show(irqmask, IRQMASK);
cp_reg_write(irqmask, IRQMASK);
cp_reg_show(irqsense, IRQSENSE);
cp_reg_write(irqsense, IRQSENSE);
cp_reg_show(irqedge, IRQEDGE);
cp_reg_write(irqedge, IRQEDGE);
cp_reg_show(irqboth, IRQBOTH);
cp_reg_write(irqboth, IRQBOTH);
cp_reg_show(irqstat_raw, IRQSTAT_RAW);
cp_reg_write(irqstat_raw, IRQSTAT_RAW);
cp_reg_show(irqstat, IRQSTAT);
cp_reg_write(irqstat, IRQSTAT);
cp_reg_show(dac0, DAC_A);
cp_reg_write(dac0, DAC_A);
cp_reg_show(dac1, DAC_B);
cp_reg_write(dac1, DAC_B);
cp_reg_show(adc0, ADC_0);
cp_reg_show(adc1, ADC_1);
cp_reg_show(adc2, ADC_2);
cp_reg_show(adc3, ADC_3);
cp_reg_show(irqstat2_raw, IRQSTAT2_RAW);
cp_reg_write(irqstat2_raw, IRQSTAT2_RAW);
cp_reg_show(irqmask2, IRQMASK2);
cp_reg_write(irqmask2, IRQMASK2);
cp_reg_show(irqstat2, IRQSTAT2);
cp_reg_write(irqstat2, IRQSTAT2);
cp_reg_show(ex0_31, EX_0_31);
cp_reg_write(ex0_31, EX_0_31);
cp_reg_show(ex32_41, EX_32_41);
cp_reg_write(ex32_41, EX_32_41);
cp_reg_show(ex0_31_dir, EX_0_31_DIR);
cp_reg_write(ex0_31_dir, EX_0_31_DIR);
cp_reg_show(ex32_41_dir, EX_32_41_DIR);
cp_reg_write(ex32_41_dir, EX_32_41_DIR);

#define attr_r(reg)	\
__ATTR(cp_##reg, S_IRUGO, cp_##reg##_show, NULL)

#define attr_w(reg)	\
__ATTR(cp_##reg, S_IWUSR, NULL, cp_##reg##_write)

#define attr_rw(reg)	\
__ATTR(cp_##reg, S_IRUGO | S_IWUSR, cp_##reg##_show, cp_##reg##_write)

static struct device_attribute cp_dev_attrs[] = {
	attr_rw(dout),
	attr_w(dout_set),
	attr_w(dout_clr),
	attr_r(din),
	attr_rw(debounce),
	attr_rw(irqmask),
	attr_rw(irqsense),
	attr_rw(irqedge),
	attr_rw(irqboth),
	attr_rw(irqstat_raw),
	attr_rw(irqstat),
	attr_rw(dac0),
	attr_rw(dac1),
	attr_r(adc0),
	attr_r(adc1),
	attr_r(adc2),
	attr_r(adc3),
	attr_rw(irqstat2_raw),
	attr_rw(irqmask2),
	attr_rw(irqstat2),
	attr_rw(ex0_31),
	attr_rw(ex32_41),
	attr_rw(ex0_31_dir),
	attr_rw(ex32_41_dir),
	__ATTR_NULL
};

static const struct file_operations cp_fops = {
	.owner = THIS_MODULE,
	.llseek = no_llseek,
	.read = cp_read,
	.write = cp_write,
	.ioctl = cp_ioctl,
	.mmap = cp_mmap,
	.open = cp_open,
	.release = cp_release,
	.fasync = cp_fasync,
};

static int cp_probe(struct pci_dev *dev, const struct pci_device_id *id)
{
	static int	brd_num = -1;
	static int	brd_idx = -1;
	struct cp_board	*brd;
	unsigned long	phys_io;
	void __iomem	*virt_io;

	brd_num++;

	printk(KERN_INFO CP_MOD "KSI Labs CENTIPEDE Board %d found\n", brd_num);

	if (brd_idx >= (MAX_BOARDS - 1)) {
		printk(KERN_ERR CP_MOD "Skipping--only %d boards allowed\n", MAX_BOARDS);
		return -ENODEV;
	}

	if (pci_enable_device(dev)) {
		printk(KERN_ERR CP_MOD "Failed to enable PCI device for board %d\n", brd_num);
		return -EIO;
	}

	phys_io = pci_resource_start(dev, 0);
	if (!phys_io) {
		printk(KERN_ERR CP_MOD "PCI region 0 not available for board %d\n", brd_num);
		goto err_out;
	}

	if (!request_mem_region(phys_io, IOMEM_LEN, "centipede")) {
		printk(KERN_ERR CP_MOD "Resource 0x%x @ 0x%lx unavailable\n", IOMEM_LEN, phys_io);
		goto err_out;
	}

	virt_io = ioremap_nocache(pci_resource_start(dev, 0), IOMEM_LEN);
	if (!virt_io) {
		printk(KERN_ERR CP_MOD "Failed to ioremap PCI memory space at 0x%lx\n", phys_io);
		goto err_mem_reg;
	}

	brd_idx++;
	
	brd = kzalloc(sizeof(struct cp_board), GFP_KERNEL);
	if (!brd) {
		printk(KERN_ERR CP_MOD "Failed to allocate cp_board structure\n");
		goto err_idx_dec;
	}

	spin_lock_init(&brd->mutex);

	/* Clear and disable interrupts */
	writel(0x0L, virt_io + CP_IRQMASK_OFF);
	writel(0x0L, virt_io + CP_IRQMASK2_OFF);
	writel(0xffffffff, virt_io + CP_IRQSTAT_RAW_OFF);
	writel(0x7fffffff, virt_io + CP_IRQSTAT2_RAW_OFF);
	(void) readl(virt_io + CP_IRQMASK_OFF);

	brd->phys_io = phys_io;
	brd->virt_io = virt_io;
	brd->irq = dev->irq;
	cdev_init(&brd->cdev, &cp_fops);
	brd->cdev.owner = THIS_MODULE;
	brd->cdev.ops = &cp_fops;

	if (request_irq(dev->irq, &cp_interrupt, IRQF_SHARED, "centipede", brd)) {
		printk(KERN_ERR CP_MOD "Could not allocate IRQ %d for CENTIPEDE PCI device\n", dev->irq);
		goto err_kfree;
	}

	if (cdev_add(&brd->cdev, MKDEV(cp_major, brd_idx), 1)) {
		printk(KERN_ERR CP_MOD "Error adding centipede%d device", brd_idx);
		goto err_all;
	}

	pci_set_drvdata(dev, brd);

	snprintf(brd->cp_name, 12, "centipede%d", brd_idx);
	brd->cp_dev = device_create(cp_class, NULL,
		MKDEV(cp_major, brd_idx), brd, brd->cp_name);
	if(!brd->cp_dev) {
		printk(KERN_ERR CP_MOD "device_create() failed for %s", brd->cp_name);
		cdev_del(&brd->cdev);
		goto err_all;
	}

	cp_brds[brd_idx] = brd;

	return 0;

err_all:
	free_irq(cp_brds[brd_idx]->irq, cp_brds[brd_idx]);
err_kfree:
	kfree(cp_brds[brd_idx]);
	cp_brds[brd_idx] = NULL;
err_idx_dec:
	brd_idx--;
	iounmap(virt_io);
err_mem_reg:
	release_mem_region(phys_io, IOMEM_LEN);
err_out:
	pci_disable_device(dev);
	return -ENODEV;
}

static void cp_remove(struct pci_dev *dev)
{
	struct cp_board	*brd = pci_get_drvdata(dev);
	
	if (!brd) return;

	cdev_del(&brd->cdev);
	device_del(brd->cp_dev);

	/* Clear and disable interrupts */
	writel(0x0L, brd->virt_io + CP_IRQMASK_OFF);
	writel(0x0L, brd->virt_io + CP_IRQMASK2_OFF);
	writel(0xffffffff, brd->virt_io + CP_IRQSTAT_RAW_OFF);
	writel(0x7fffffff, brd->virt_io + CP_IRQSTAT2_RAW_OFF);
	(void) readl(brd->virt_io + CP_IRQMASK_OFF);
	
	free_irq(brd->irq, brd);
	iounmap(brd->virt_io);
	release_mem_region(brd->phys_io, IOMEM_LEN);
	kfree(brd);
	brd = NULL;
	pci_disable_device(dev);
}

static struct pci_driver cp_pci_driver = {
	.name		=	"centipede",
	.id_table	=	cp_pci_tbl,
	.probe		=	cp_probe,
	.remove		=	cp_remove,
};

static int __init cp_init(void)
{
	int	i,ret;
	dev_t	devno;

	printk(KERN_INFO "KSI Labs CENTIPEDE driver, v.1.0");

	cp_class = class_create(THIS_MODULE, "centipede");
	if (IS_ERR(cp_class)) {
		printk(KERN_ERR CP_MOD "class_create() failed for cp_class");
		return -ENODEV;
	}
	cp_class->dev_attrs = cp_dev_attrs;

	ret = alloc_chrdev_region(&devno, 0, MAX_BOARDS, "centipede");
	if (ret < 0) {
		printk(KERN_ERR CP_MOD "Can't get chrdev_region\n");
		class_destroy(cp_class);
		return ret;
	}

	cp_major = MAJOR(devno);

	for (i = 0; i < MAX_BOARDS; i++)
		cp_brds[i] = NULL;

	return pci_register_driver(&cp_pci_driver);
}

static void __exit cp_exit(void)
{
	pci_unregister_driver(&cp_pci_driver);

	if (cp_class)
		class_destroy(cp_class);

	if (cp_major)
		unregister_chrdev_region(MKDEV(cp_major, 0), MAX_BOARDS);
}

module_init(cp_init);
module_exit(cp_exit);
